
class MultiValueDictKeyError(Exception):
    def __init__(self, message="Missing Key in dict"):
        self.message = message
        super().__init__(self.message)


class TelnetUnexpectedResponse(Exception):
    def __init__(self, message="Unexpected response from Jasmin"):
        self.message = message
        super().__init__(self.message)


class TelnetConnectionTimeout(Exception):
    def __init__(self, message="Connection to jcli timed out"):
        self.message = message
        super().__init__(self.message)


class TelnetLoginFailed(Exception):
    def __init__(self, message="Jasmin login failed"):
        self.message = message
        super().__init__(self.message)


class CanNotModifyError(Exception):
    def __init__(self, message="Can not modify a key"):
        self.message = message
        super().__init__(self.message)


class JasminSyntaxError(Exception):
    def __init__(self, message="Can not modify a key"):
        self.message = message
        super().__init__(self.message)


class JasminError(Exception):
    def __init__(self, message="Jasmin error"):
        self.message = message
        super().__init__(self.message)


class UnknownError(Exception):
    def __init__(self, message="object not known"):
        self.message = message
        super().__init__(self.message)


class MissingKeyError(Exception):
    def __init__(self, message="A mandatory key is missing"):
        self.message = message
        super().__init__(self.message)


class MutipleValuesRequiredKeyError(Exception):
    def __init__(self, message="Multiple values are required fro this key"):
        self.message = message
        super().__init__(self.message)


class ActionFailed(Exception):
    def __init__(self, message="Action failed"):
        self.message = message
        super().__init__(self.message)


class ObjectNotFoundError(Exception):
    def __init__(self, message="Object not found"):
        self.message = message
        super().__init__(self.message)



from .conn import TelnetConnection
from .basicmanager import BasicManager, JasminSyntaxError, UnknownError


class User(BasicManager):
    "Users for managing *Jasmin* users"

    def __init__(self, conn: TelnetConnection):
        super().__init__(conn=conn, module="user", lookup_field="uid")
        self.statusHeaderID = "User id"
    
    def partial_update(self, data: dict[str, str | int | float | bool], sub_id: str) -> bool:
        """Update some SubModule attributes
        """
        if not ((type(data) is dict) and (len(data) >= 1)):
            raise JasminSyntaxError("updates should be a dict")

        processed_data = self._split_dict_cols(data)
        data = [x for x in processed_data if x[0]
                not in self._getGuardedAttributes()]

        if not ((type(data) is list) and (len(data) >= 1)):
            raise JasminSyntaxError(
                "updates should contains some keys which can be updated")

        self.telnet.sendline(f"{self.module} -u {sub_id}")
        matched_index = self.telnet.expect([
            r".*Updating (.*)" + self.INTERACTIVE_PROMPT,
            r".*Unknown .+: (.*)" + self.STANDARD_PROMPT,
            r".+(.*)(" + self.INTERACTIVE_PROMPT +
            "|" + self.STANDARD_PROMPT + ")",
        ])
        if matched_index == 1:
            raise UnknownError(f"Unknown {self.module}: {sub_id}")
        if matched_index != 0:
            raise UnknownError(self.telnet.match.group(0))
        for update in data:
            if not ((type(update) is list) and (len(update) >= 1)):
                raise JasminSyntaxError("Not a list: %s" % update)
            self.telnet.sendline(" ".join([x for x in update]))
            matched_index = self.telnet.expect([
                r".*(Unknown .+ key:.*)" + self.INTERACTIVE_PROMPT,
                r".*(Error:.*)" + self.STANDARD_PROMPT,
                r".*" + self.INTERACTIVE_PROMPT,
                r".+(.*)(" + self.INTERACTIVE_PROMPT +
                "|" + self.STANDARD_PROMPT + ")",
            ])
            if matched_index != 2:
                raise JasminSyntaxError(
                    " ".join(self.telnet.match.group(1).split()))
        self.telnet.sendline("ok")
        ok_index = self.telnet.expect([
            r"(.*)" + self.INTERACTIVE_PROMPT,
            r".*" + self.STANDARD_PROMPT,
        ])
        if ok_index == 0:
            raise JasminSyntaxError(
                " ".join(self.telnet.match.group(1).split()))
        self._persist()
        return True

    def _parseSingleStatus(self, status_list: list) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        return {
            self.statusHeaderID: status_list[0].lstrip("!"),
            "Group id": status_list[1].lstrip("!"),
            "Username": status_list[2],
            "Balance MT": status_list[3],
            "Balance SMS": status_list[4],
            "Throughput": status_list[5],
            "status": not status_list[0][0] == "!" and not status_list[1][0] == "!"
        }

    def _parseSingle(self, sub_module: dict[str, str | int | float | bool]) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        allSubModulesStatus = self.allSubModulesStatus if hasattr(
            self, "allSubModulesStatus") else self.listStatus()
        if not any(d[self.statusHeaderID] == sub_module[self.lookup_field] for d in allSubModulesStatus):
            sub_module["status"] = False
            return sub_module
        userStatus = [u for u in allSubModulesStatus if u[self.statusHeaderID]
                      == sub_module[self.lookup_field]][0]
        sub_module["status"] = userStatus["status"]
        return sub_module

    def enable(self, sub_id: str) -> bool | str:
        """Enable a user. One parameter required, the user identifier (a string)
        """
        return self._simple_submodule_action(action="e", sub_id=sub_id)

    def disable(self, sub_id: str) -> bool | str:
        """Disable a user. One parameter required, the user identifier (a string)
        """
        return self._simple_submodule_action(action="d", sub_id=sub_id)

    def smpp_unbind(self, sub_id: str) -> bool | str:
        """Unbind user from smpp server
        """
        return self._simple_submodule_action(action="-smpp-unbind", sub_id=sub_id)

    def smpp_ban(self, sub_id: str) -> bool | str:
        """Unbind and ban user from smpp server
        """
        return self._simple_submodule_action(action="-smpp-ban", sub_id=sub_id)

    def _getRequiredAttributes(self, data: dict[str, str | int | float | bool]) -> list[str]:
        """ Required for create, can use data:dict to check if some keys depends on others """
        prioritAttributes = self._getPrioritAttributes()
        requiredAttributes = ["username", "password"]
        return prioritAttributes + list(set(requiredAttributes) - set(prioritAttributes))

    def _getPrioritAttributes(self) -> list[str]:
        """ Always sent first """
        return ["uid", "gid"]

    def _getGuardedAttributes(self) -> list[str]:
        """ NEVER send for Update """
        return ["username", "uid"]



from .conn import TelnetConnection
from .basicmanager import BasicManager, ObjectNotFoundError


class Group(BasicManager):
    "Users for managing *Jasmin* users"

    def __init__(self, conn: TelnetConnection):
        super().__init__(conn=conn, module="group", lookup_field="gid")
        self.statusHeaderID = "Group id"

    def _parseSingleStatus(self, status_list: list) -> dict[str, str | int | float | bool]:
        """ Groups for managing *Jasmin* user groups (*not* Django auth groups) """
        return {
            self.statusHeaderID: status_list[0].lstrip("!"),
            "status": not status_list[0][0] == "!"
        }

    def get_one(self, sub_id: str) -> dict[str, str | int | float | bool]:
        """Gets a single SubModule data """
        allSubModulesStatus = self.allSubModulesStatus if hasattr(
            self, "allSubModulesStatus") else self.listStatus()
        if not any(d[self.statusHeaderID] == sub_id for d in allSubModulesStatus):
            raise ObjectNotFoundError(f"Unknown {self.module}: {sub_id}")
        return [u for u in allSubModulesStatus if u[self.statusHeaderID] == sub_id][0]

    def enable(self, sub_id: str) -> bool | str:
        """Enable a group. One parameter required, the group identifier (a string)
        """
        return self._simple_submodule_action(action="e", sub_id=sub_id)

    def disable(self, sub_id: str) -> bool | str:
        """Disable a group. One parameter required, the group identifier (a string)
        """
        return self._simple_submodule_action(action="d", sub_id=sub_id)

    def _getRequiredAttributes(self, data: dict) -> list[str]:
        """ Required for create, can use data:dict to check if some keys depends on others """
        prioritAttributes = self._getPrioritAttributes()
        requiredAttributes = []
        return prioritAttributes + list(set(requiredAttributes) - set(prioritAttributes))

    def _getPrioritAttributes(self) -> list[str]:
        """ Always sent first """
        return ["gid"]

    def _getGuardedAttributes(self) -> list[str]:
        """ NEVER send for Update """
        return ["gid"]



from .conn import TelnetConnection
from .basicmanager import BasicManager, ObjectNotFoundError
import re


class MOInterceptor(BasicManager):
    "MOInterceptor for managing MO Interceptors"

    def __init__(self, conn: TelnetConnection):
        super().__init__(conn=conn, module="mointerceptor", lookup_field="order")
        self.statusHeaderID = "Order"

    def _parseSingleStatus(self, status_list: list) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        return {
            self.statusHeaderID: status_list[0],
            "Type": status_list[1],
            "Script": re.findall("<MOIS.*?>", " ".join(status_list)),
            "Filter(s)": re.findall("<(?!MOIS).*?>", " ".join(status_list))
        }

    def get_one(self, sub_id: str) -> dict[str, str | int | float | bool]:
        """Gets a single SubModule data """
        allSubModulesStatus = self.allSubModulesStatus if hasattr(
            self, "allSubModulesStatus") else self.listStatus()
        if not any(d[self.statusHeaderID] == sub_id for d in allSubModulesStatus):
            raise ObjectNotFoundError(f"Unknown {self.module}: {sub_id}")
        return [u for u in allSubModulesStatus if u[self.statusHeaderID] == sub_id][0]

    def flush(self) -> bool | str:
        "Flush entire MOInterceptor table"
        self._simple_submodule_action(action="f", sub_id="")

    def _getRequiredAttributes(self, data: dict[str, str | int | float | bool]) -> list[str]:
        """ Required for create, can use data:dict to check if some keys depends on others """
        prioritAttributes = self._getPrioritAttributes()
        requiredAttributes = []
        if "type" not in data:
            pass

        elif data["type"].lower() == "staticmointerceptor":
            requiredAttributes.extend(["filters", "script", "order"])

        elif data["type"].lower() == "defaultinterceptor":
            requiredAttributes.extend(["script"])

        return prioritAttributes + list(set(requiredAttributes) - set(prioritAttributes))

    def _getPrioritAttributes(self) -> list[str]:
        """ Always sent first """
        return ["type"]

    def _getGuardedAttributes(self) -> list[str]:
        """ NEVER send for Update """
        return ["order"]



from pexpect import spawn
import pexpect

from .exceptions import TelnetUnexpectedResponse, TelnetConnectionTimeout, TelnetLoginFailed


class TelnetConnection(object):
    def __init__(self, host: str = "127.0.0.1", port: int = 8990, timeout: int = 10, auth:bool=True, username: str = "jcliadmin", password: str = "jclipwd", standard_prompt: str = "jcli : ", interactive_prompt: str = "> "):
        self.STANDARD_PROMPT = standard_prompt
        self.INTERACTIVE_PROMPT = interactive_prompt
        try:
            telnet: spawn = pexpect.spawn(f"telnet {host} {port}", timeout=timeout)
            if auth:
                telnet.expect_exact("Username: ")
                telnet.sendline(username)
                telnet.expect_exact("Password: ")
                telnet.sendline(password)
        except pexpect.EOF:
            raise TelnetUnexpectedResponse
        except pexpect.TIMEOUT:
            raise TelnetConnectionTimeout
        try:
            telnet.expect_exact(self.STANDARD_PROMPT)
        except pexpect.EOF:
            raise TelnetLoginFailed
        else:
            self.telnet = telnet
            self.telnet.setwinsize(200, 4000)
            self.telnet.maxread = 4000

    def __del__(self):
        "Make sure telnet connection is closed when unleashing response back to client"
        try:
            self.telnet.sendline("quit")
        except pexpect.ExceptionPexpect:
            self.telnet.kill(9)



from .conn import TelnetConnection
from .basicmanager import BasicManager, ObjectNotFoundError


class Filter(BasicManager):
    "Filters Class"

    def __init__(self, conn: TelnetConnection):
        super().__init__(conn=conn, module="filter", lookup_field="fid")
        self.statusHeaderID = "Filter id"

    def _parseSingleStatus(self, status_list: list) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        return {
            self.statusHeaderID: status_list[0],
            "Type": status_list[1],
            "Routes": status_list[2] + " " + status_list[3],
            "Description": " ".join(status_list[4:])
        }

    def get_one(self, sub_id: str) -> dict[str, str | int | float | bool]:
        """Gets a single SubModule data """
        allSubModulesStatus = self.allSubModulesStatus if hasattr(
            self, "allSubModulesStatus") else self.listStatus()
        if not any(d[self.statusHeaderID] == sub_id for d in allSubModulesStatus):
            raise ObjectNotFoundError(f"Unknown {self.module}: {sub_id}")
        return [u for u in allSubModulesStatus if u[self.statusHeaderID] == sub_id][0]

    def _getRequiredAttributes(self, data: dict[str, str | int | float | bool]) -> list[str]:
        """ Required for create, can use data:dict to check if some keys depends on others """
        prioritAttributes = self._getPrioritAttributes()
        requiredAttributes = ["fid"]
        if "type" not in data or data["type"].lower() == "transparentfilter":
            pass

        elif data["type"].lower() == "connectorfilter":
            requiredAttributes.extend(["cid"])

        elif data["type"].lower() == "userfilter":
            requiredAttributes.extend(["uid"])

        elif data["type"].lower() == "groupfilter":
            requiredAttributes.extend(["gid"])

        elif data["type"].lower() == "sourceaddrfilter":
            requiredAttributes.extend(["source_addr"])

        elif data["type"].lower() == "destinationaddrfilter":
            requiredAttributes.extend(["destination_addr"])

        elif data["type"].lower() == "shortmessagefilter":
            requiredAttributes.extend(["short_message"])

        elif data["type"].lower() == "dateintervalfilter":
            requiredAttributes.extend(["dateInterval"])

        elif data["type"].lower() == "timeintervalfilter":
            requiredAttributes.extend(["timeInterval"])

        elif data["type"].lower() == "tagfilter":
            requiredAttributes.extend(["tag"])

        elif data["type"].lower() == "evalpyfilter":
            requiredAttributes.extend(["pyCode"])

        return prioritAttributes + list(set(requiredAttributes) - set(prioritAttributes))

    def _getPrioritAttributes(self) -> list[str]:
        """ Always sent first """
        return ["type"]

    def _getGuardedAttributes(self) -> list[str]:
        """ NEVER send for Update """
        return ["fid"]



from .conn import TelnetConnection
from .filter import Filter as Telnet_Filter
from .group import Group as Telnet_Group
from .httpccm import HTTPCCM as Telnet_HTTPCCM
from .mointerceptor import MOInterceptor as Telnet_MOInterceptor
from .morouter import MORouter as Telnet_MORouter
from .mtinterceptor import MTInterceptor as Telnet_MTInterceptor
from .mtrouter import MTRouter as Telnet_MTRouter
from .smppccm import SMPPCCM as Telnet_SMPPCCM
from .user import User as Telnet_User
import threading

class Proxy:
    def __init__(self, host: str = "127.0.0.1", port: int = 8990, timeout: int = 10, auth: bool = True, username: str = "jcliadmin", password: str = "jclipwd", standard_prompt: str = "jcli : ", interactive_prompt: str = "> ", log_status: bool = False, script_log: str = "JasminTelnet-Proxy-> "):
        """ Constructor """
        self.telnet_host = host
        self.telnet_port = port
        self.telnet_timeout = timeout
        self.telnet_auth = auth
        self.telnet_username = username
        self.telnet_password = password
        self.telnet_standard_prompt = standard_prompt
        self.telnet_interactive_prompt = interactive_prompt
        self.log_status = log_status
        self.script_log = script_log

    def getTelnetConn(self) -> TelnetConnection:
        return TelnetConnection(
            host=self.telnet_host,
            port=self.telnet_port,
            timeout=self.telnet_timeout,
            auth=self.telnet_auth,
            username=self.telnet_username,
            password=self.telnet_password,
            standard_prompt=self.telnet_standard_prompt,
            interactive_prompt=self.telnet_interactive_prompt,
        )

    def log(self, log: str):
        if self.log_status is True:
            print("{:<65}".format(f"{self.script_log}{log}")+"***", flush=True)

    def syncAll(self, collection_data: dict[str, dict[str, dict[str, str | int | float | bool]]]):
        """ Excute Sync all modules function """
        if not isinstance(collection_data, dict):
            collection_data = {}
        if "smppccm" not in collection_data:
            collection_data["smppccm"] = False
        if "httpccm" not in collection_data:
            collection_data["httpccm"] = False
        if "group" not in collection_data:
            collection_data["group"] = False
        if "filter" not in collection_data:
            collection_data["filter"] = False
        if "mointerceptor" not in collection_data:
            collection_data["mointerceptor"] = False
        if "mtinterceptor" not in collection_data:
            collection_data["mtinterceptor"] = False
        if "user" not in collection_data:
            collection_data["user"] = False
        if "morouter" not in collection_data:
            collection_data["morouter"] = False
        if "mtrouter" not in collection_data:
            collection_data["mtrouter"] = False
        self.syncCollection(collection_data)

    def syncCollection(self, collection_data: dict[str, dict[str, dict[str, str | int | float | bool]]]):
        """ Excute Sync a collection of modules with threads function """
        jobs_levels: list[list[threading.Thread], list[threading.Thread], list[threading.Thread],
                          list[threading.Thread], list[threading.Thread]] = [[], [], [], [], []]

        if "group" in collection_data:
            jobs_levels[0].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "group",
            }))
        if "smppccm" in collection_data:
            jobs_levels[0].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "smppccm",
            }))
        if "httpccm" in collection_data:
            jobs_levels[0].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "httpccm",
            }))
        if "user" in collection_data:
            jobs_levels[1].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "user",
            }))
        if "filter" in collection_data:
            jobs_levels[2].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "filter",
            }))
        if "mointerceptor" in collection_data:
            jobs_levels[3].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "mointerceptor",
            }))
        if "mtinterceptor" in collection_data:
            jobs_levels[3].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "mtinterceptor",
            }))
        if "morouter" in collection_data:
            jobs_levels[4].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "morouter",
            }))
        if "mtrouter" in collection_data:
            jobs_levels[4].append(threading.Thread(target=self.syncOneFromCollection, kwargs={
                "collection_data": collection_data,
                "module": "mtrouter",
            }))

        # Loop all jobs levels
        for jobs_level in jobs_levels:
            # Start all jobs in this level
            for job in jobs_level:
                job.start()
            # Wait for all jobs in this level to finish
            for job in jobs_level:
                try:
                    job.join()
                except KeyboardInterrupt:
                    pass
        del jobs_levels

    def syncOneFromCollection(self, module: str, collection_data: dict[str, dict[str, dict[str, str | int | float | bool]]]):
        """ Excute Sync a given module from a collection function """
        self.sync(module=module, sub_modules_data=collection_data[module])

    def sync(self, module: str, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        """ Excute Sync a given module function """
        telnet = self.getTelnetConn()

        if module == "smppccm":
            self.log("***          Sync: Smppccm module.")
            Smppccm(telnet_conn=telnet).sync(sub_modules_data=sub_modules_data)
        if module == "httpccm":
            self.log("***          Sync: Httpccm module.")
            Httpccm(telnet_conn=telnet).sync(sub_modules_data=sub_modules_data)
        if module == "group":
            self.log("***          Sync: Group module.")
            Group(telnet_conn=telnet).sync(sub_modules_data=sub_modules_data)
        if module == "user":
            self.log("***          Sync: User module.")
            User(telnet_conn=telnet).sync(sub_modules_data=sub_modules_data)
        if module == "filter":
            self.log("***          Sync: Filter module.")
            Filter(telnet_conn=telnet).sync(sub_modules_data=sub_modules_data)
        if module == "mointerceptor":
            self.log("***          Sync: MoInterceptor module.")
            MOInterceptor(telnet_conn=telnet).sync(
                sub_modules_data=sub_modules_data)
        if module == "morouter":
            self.log("***          Sync: MoRouter module.")
            Morouter(telnet_conn=telnet).sync(
                sub_modules_data=sub_modules_data)
        if module == "mtinterceptor":
            self.log("***          Sync: MtInterceptor module.")
            MTInterceptor(telnet_conn=telnet).sync(
                sub_modules_data=sub_modules_data)
        if module == "mtrouter":
            self.log("***          Sync: MtRouter module.")
            Mtrouter(telnet_conn=telnet).sync(
                sub_modules_data=sub_modules_data)

    def add(self, module: str, sub_id: str, options: dict[str, str | int | float | bool]):
        """ Excute Add to a module function """
        telnet = self.getTelnetConn()

        if module == "smppccm":
            self.log("***          Add new: Smppccm module.")
            Smppccm(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "httpccm":
            self.log("***          Add new: Httpccm module.")
            Httpccm(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "group":
            self.log("***          Add new: Group module.")
            Group(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "user":
            self.log("***          Add new: User module.")
            User(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "filter":
            self.log("***          Add new: Filter module.")
            Filter(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "mointerceptor":
            self.log("***          Add new: MoInterceptor module.")
            MOInterceptor(telnet_conn=telnet).add(
                sub_id=sub_id, options=options)
        if module == "morouter":
            self.log("***          Add new: MoRouter module.")
            Morouter(telnet_conn=telnet).add(sub_id=sub_id, options=options)
        if module == "mtinterceptor":
            self.log("***          Add new: MtInterceptor module.")
            MTInterceptor(telnet_conn=telnet).add(
                sub_id=sub_id, options=options)
        if module == "mtrouter":
            self.log("***          Add new: MtRouter module.")
            Mtrouter(telnet_conn=telnet).add(sub_id=sub_id, options=options)

    def edit(self, module: str, sub_id: str, options: dict[str, str | int | float | bool]):
        """ Excute Edit a module function """
        telnet = self.getTelnetConn()

        if module == "smppccm":
            self.log("***          Update: Smppccm module.")
            Smppccm(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "httpccm":
            self.log("***          Update: Httpccm module.")
            Httpccm(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "group":
            self.log("***          Update: Group module.")
            Group(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "user":
            self.log("***          Update: User module.")
            User(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "filter":
            self.log("***          Update: Filter module.")
            Filter(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "mointerceptor":
            self.log("***          Update: MoInterceptor module.")
            MOInterceptor(telnet_conn=telnet).update(
                sub_id=sub_id, options=options)
        if module == "morouter":
            self.log("***          Update: MoRouter module.")
            Morouter(telnet_conn=telnet).update(sub_id=sub_id, options=options)
        if module == "mtinterceptor":
            self.log("***          Update: MtInterceptor module.")
            MTInterceptor(telnet_conn=telnet).update(
                sub_id=sub_id, options=options)
        if module == "mtrouter":
            self.log("***          Update: MtRouter module.")
            Mtrouter(telnet_conn=telnet).update(sub_id=sub_id, options=options)

    def remove(self, module: str, sub_id: str):
        """ Excute Remove from a module function """
        telnet = self.getTelnetConn()

        if module == "smppccm":
            self.log("***          Remove: Smppccm module.")
            Smppccm(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "httpccm":
            self.log("***          Remove: Httpccm module.")
            Httpccm(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "group":
            self.log("***          Remove: Group module.")
            Group(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "user":
            self.log("***          Remove: User module.")
            User(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "filter":
            self.log("***          Remove: Filter module.")
            Filter(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "mointerceptor":
            self.log("***          Remove: MoInterceptor module.")
            MOInterceptor(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "morouter":
            self.log("***          Remove: MoRouter module.")
            Morouter(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "mtinterceptor":
            self.log("***          Remove: MtInterceptor module.")
            MTInterceptor(telnet_conn=telnet).remove(sub_id=sub_id)
        if module == "mtrouter":
            self.log("***          Remove: MtRouter module.")
            Mtrouter(telnet_conn=telnet).remove(sub_id=sub_id)


class Filter:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_Filter(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class Group:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_Group(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        processed_data = {key: options[key] for key in set(
            list(options.keys())) - set(["status"])}
        self.manager.create(data=processed_data)
        if "status" in options.keys():
            if options["status"]:
                self.manager.enable(sub_id=sub_id)
            else:
                self.manager.disable(sub_id=sub_id)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class Httpccm:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_HTTPCCM(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class MOInterceptor:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_MOInterceptor(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class Morouter:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_MORouter(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class MTInterceptor:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_MTInterceptor(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class Mtrouter:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_MTRouter(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.manager.create(data=options)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        self.add(sub_id=sub_id, options=options)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class Smppccm:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_SMPPCCM(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        processed_data = {key: options[key] for key in set(
            list(options.keys())) - set(["status"])}
        self.manager.create(data=processed_data)
        if "status" in options.keys():
            if options["status"]:
                self.manager.start(sub_id=sub_id)
            else:
                self.manager.stop(sub_id=sub_id)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        processed_data = {key: options[key] for key in set(
            list(options.keys())) - set(["status"])}
        self.manager.partial_update(data=processed_data, sub_id=sub_id)
        if "status" in options.keys():
            if options["status"]:
                self.manager.start(sub_id=sub_id)
            else:
                self.manager.stop(sub_id=sub_id)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)


class User:
    def __init__(self, telnet_conn: TelnetConnection):
        self.manager = Telnet_User(conn=telnet_conn)

    def sync(self, sub_modules_data: dict[str, dict[str, str | int | float | bool]]):
        synced_ids: list[str] = []
        manager_ids: list[str] = [sub_status[self.manager.statusHeaderID]
                                  for sub_status in self.manager.listStatus()]
        for sub_id, options in sub_modules_data.items():
            synced_ids.append(sub_id)
            if sub_id in manager_ids:
                self.update(sub_id=sub_id, options=options)
            else:
                self.add(sub_id=sub_id, options=options)
                manager_ids.append(sub_id)

        for manager_id in manager_ids:
            if manager_id not in synced_ids:
                self.remove(sub_id=manager_id)

    def add(self, sub_id: str, options: dict[str, str | int | float | bool]):
        processed_data = {key: options[key] for key in set(
            list(options.keys())) - set(["status"])}
        self.manager.create(data=processed_data)
        if "status" in options.keys():
            if options["status"]:
                self.manager.enable(sub_id=sub_id)
            else:
                self.manager.disable(sub_id=sub_id)

    def update(self, sub_id: str, options: dict[str, str | int | float | bool]):
        processed_data = {key: options[key] for key in set(
            list(options.keys())) - set(["status"])}
        self.manager.partial_update(data=processed_data, sub_id=sub_id)
        if "status" in options.keys():
            if options["status"]:
                self.manager.enable(sub_id=sub_id)
            else:
                self.manager.disable(sub_id=sub_id)

    def remove(self, sub_id):
        self.manager.destroy(sub_id=sub_id)



from .conn import TelnetConnection
from .exceptions import (CanNotModifyError,
                         JasminSyntaxError, JasminError,
                         UnknownError, MissingKeyError,
                         ObjectNotFoundError)


class BasicManager(object):
    "Basic Manager for managing *Jasmin*"

    def __init__(self, conn: TelnetConnection, module: str, lookup_field: str, defaultPersist: bool = True):
        self.conn = conn
        self.telnet = conn.telnet
        self.STANDARD_PROMPT = conn.STANDARD_PROMPT
        self.INTERACTIVE_PROMPT = conn.INTERACTIVE_PROMPT
        self.module = module
        self.lookup_field = lookup_field
        self.defaultPersist = defaultPersist

    def create(self, data: dict[str, str | int | float | bool]) -> bool:
        """Create a SubModule.
        """
        if set(self._getRequiredAttributes(data=data)) > set(data):
            raise MissingKeyError(
                f"Missing parameter: {self._getRequiredAttributes(data=data)} required")

        data = self._prepareAttributes(data)
        self.telnet.sendline(f"{self.module} -a")
        self.telnet.expect(r"Adding a new (.+)\n" +
                           self.INTERACTIVE_PROMPT)
        self._set_ikeys(
            keys2vals=data, priorityKeys=self._getPrioritAttributes())
        self._persist()
        return True

    def listAll(self) -> list[dict[str, str | int | float | bool]]:
        "List SubModules. No parameters"
        self.telnet.sendline(f"{self.module} -l")
        self.telnet.expect([r"(.+)\n" + self.STANDARD_PROMPT])
        result = self.telnet.match.group(0).strip()
        if len(result) < 3:
            return []

        results = [l for l in result.splitlines() if l]

        self.allSubModulesStatus: list[dict] = [self._parseSingleStatus(status_list=str(
            u, "utf-8").lstrip("#").lstrip().split(None)) for u in results[2:-2]]

        sub_modules = [self.get_one(sub_id=str(u, "utf-8").lstrip("#").lstrip().split(None, 1)[0].lstrip("!"))
                       for u in results[2:-2]]
        delattr(self, "allSubModulesStatus")
        return [
            # return sub_modules skipping None (== nonexistent sub_module)
            sub_module for sub_module in sub_modules if sub_modules
        ]

    def listStatus(self) -> list[dict[str, str | int | float | bool]]:
        "List SubModules Status. No parameters"
        self.telnet.sendline(f"{self.module} -l")
        self.telnet.expect([r"(.+)\n" + self.STANDARD_PROMPT])
        result = self.telnet.match.group(0).strip()
        if len(result) < 3:
            return []

        results = [l for l in result.splitlines() if l]

        status_list = [self._parseSingleStatus(status_list=str(
            u, "utf-8").lstrip("#").lstrip().split(None)) for u in results[2:-2]]
        return [
            # return SubModules skipping None (== nonexistent SubModule)
            sub_module_status for sub_module_status in status_list if sub_module_status
        ]

    def get_one(self, sub_id: str) -> dict[str, str | int | float | bool]:
        """Gets a single SubModule data """
        self.telnet.sendline(f"{self.module} -s {sub_id}")
        matched_index = self.telnet.expect([
            r".+Unknown .+:.*" + self.STANDARD_PROMPT,
            r".+Usage: .+.*" + self.STANDARD_PROMPT,
            r"(.+)\n" + self.STANDARD_PROMPT,
        ])
        if matched_index != 2:
            raise ObjectNotFoundError(f"Unknown {self.module}: {sub_id}")
        result = self.telnet.match.group(1)
        sub_module = {}
        for line in [l for l in result.splitlines() if l][1:]:
            d = [str(x, "utf-8") for x in line.split() if x]
            sub_module[" ".join(d[:-1] if len(d) > 1 else d)
                       ] = d[-1] if len(d) > 1 else None
        return self._parseSingle(sub_module=sub_module)

    def destroy(self, sub_id: str):
        """Delete a SubModule. One parameter required, the SubModule identifier (a string)
        """
        return self._simple_submodule_action("r", sub_id)

    def _simple_submodule_action(self, action: str, sub_id: str) -> bool | str:
        self.telnet.sendline(f"{self.module} -{action} {sub_id}")
        matched_index = self.telnet.expect([
            r".+Successfully(.+)" + self.STANDARD_PROMPT,
            r".+Unknown .+: (.+)" + self.STANDARD_PROMPT,
            r".+(.*)" + self.STANDARD_PROMPT,
        ])
        if matched_index == 0:
            self._persist()
            return True
        elif matched_index == 1:
            raise ObjectNotFoundError(f"Unknown {self.module}: {sub_id}")
        else:
            return self.telnet.match.group(0).strip()

    def _parseSingleStatus(self, status_list: list) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        return {idx: x for idx, x in enumerate(status_list)}

    def _parseSingle(self, sub_module: dict[str, str | int | float | bool]) -> dict[str, str | int | float | bool]:
        """ Holder, each subClass should override """
        return sub_module

    def _prepareAttributes(self, data: dict[str, str | int | float | bool]) -> dict[str, str | int | float | bool]:
        """ If data needs preparing """
        return data

    def _getRequiredAttributes(self, data: dict[str, str | int | float | bool]) -> list[str]:
        """ Required for create, can use data:dict to check if some keys depends on others """
        return []

    def _getPrioritAttributes(self) -> list[str]:
        """ Always sent first """
        return []

    def _getGuardedAttributes(self) -> list[str]:
        """ NEVER Update """
        return []

    def _persist(self) -> None:
        if self.defaultPersist:
            self.telnet.sendline("persist\n")
            self.telnet.expect(r".*" + self.STANDARD_PROMPT)

    def _set_ikeys(self, keys2vals: dict, priorityKeys: list = []):
        "set multiple keys for interactive command"
        if len(priorityKeys) > 0:
            # Priority Keys first if set
            for key, val in keys2vals.items():
                if key in priorityKeys:
                    self._send_key_val(key=key, val=val)
        for key, val in keys2vals.items():
            # if priority, skip. already sent
            if key not in priorityKeys:
                self._send_key_val(key=key, val=val)
        self.telnet.sendline("ok")
        ok_index = self.telnet.expect([
            r"ok(.* syntax is invalid).*" + self.INTERACTIVE_PROMPT,
            r".*" + self.STANDARD_PROMPT,
        ])
        if ok_index == 0:
            # remove whitespace and return error
            raise JasminSyntaxError(
                " ".join(self.telnet.match.group(1).split()))
        return

    def _send_key_val(self, key: str, val: str):
        self.telnet.sendline("%s %s" % (key, val))
        matched_index = self.telnet.expect([
            r".*(Unknown .*)" + self.INTERACTIVE_PROMPT,
            r"(.*) can not be modified.*" + self.INTERACTIVE_PROMPT,
            r"(.*)" + self.INTERACTIVE_PROMPT
        ])
        result = self.telnet.match.group(1).strip()
        if matched_index == 0:
            raise UnknownError(result)
        if matched_index == 1:
            raise CanNotModifyError(result)

    def _split_dict_cols(self, keys2vals: dict[str, str | int | float | bool]) -> list[list[str]]:
        "split columns into lists"
        parsed = []
        for key, value in keys2vals.items():
            fields: list = key.split()
            fields.append(str(value))
            parsed.append(fields)
        return parsed
